/***************************************************************************
 *cr                                                                       
 *cr            (C) Copyright 1995 The Board of Trustees of the           
 *cr                        University of Illinois                       
 *cr                         All Rights Reserved                        
 *cr                                                                   
 ***************************************************************************/

/***************************************************************************
 * RCS INFORMATION:
 *
 *	$RCSfile: MoleculeRemote.C,v $
 *	$Author: billh $	$Locker:  $		$State: Exp $
 *	$Revision: 1.12 $	$Date: 95/03/24 18:50:54 $
 *
 ***************************************************************************
 * DESCRIPTION:
 *
 * A MoleculeRemote is inherited from Molecule and contains a Remote;
 * it is a Molecule which is representing a remote simulation, from which
 * it gets parameters and coordinates.
 *
 ***************************************************************************
 * REVISION HISTORY:
 *
 * $Log:	MoleculeRemote.C,v $
 * Revision 1.12  95/03/24  18:50:54  billh
 * Added copyright notice to top of file; made sure all virtual routines
 * are defined in the .C file, not in the .h file.
 * 
 * Revision 1.11  1994/11/11  16:35:17  billh
 * Now properly saves correct frames.
 *
 * Revision 1.10  94/11/10  19:26:16  billh
 * no change.
 * 
 * Revision 1.9  94/11/10  17:26:56  billh
 * Fixed stupid bugs.
 * 
 * Revision 1.8  94/11/10  17:19:57  billh
 * Added option to save every Nth transferred frame.
 * Now reads in data about patch status from mdcomm fields.
 * 
 * Revision 1.7  94/11/09  13:44:57  billh
 * Fixed to set dt properly when reading in new timesteps.
 * 
 * Revision 1.6  94/11/09  02:50:28  billh
 * Added finished support for displaying patches for a remote molecule.
 * 
 * Revision 1.5  94/11/02  01:36:40  billh
 * When receiving a timestep, must call 'get_next_ts' in Remote first, get
 * all the data, and when done, call 'done_with_ts' in Remote.
 * 
 * Revision 1.4  94/10/28  18:38:07  billh
 * Added many debug messages for when getting molecule structure.
 * 
 * Revision 1.3  1994/10/21  03:51:59  billh
 * Added initial support for patch configuration display, and fixed some bugs.
 *
 * Revision 1.2  1994/10/05  04:38:23  billh
 * Took out double backslash from text, even in comments.
 *
 * Revision 1.1  1994/10/01  11:04:31  billh
 * Initial revision
 *
 ***************************************************************************/
#ifdef ARCH_HPUX9
  static char ident[] = "@(#)$Header: /private/auto143000131/vmdsrc/vmd/billh/src/RCS/MoleculeRemote.C,v 1.12 95/03/24 18:50:54 billh Exp $";
#endif


#include "MoleculeRemote.h"
#include "Remote.h"
#include "Atom.h"
#include "DrawPatch.h"
#include "Inform.h"
#include "utilities.h"


////////////////////////////  constructor  

MoleculeRemote::MoleculeRemote(Remote *newrem, Scene *sc)
	: Molecule(newrem->host(), sc) {
  do_construct(newrem);
}

MoleculeRemote::MoleculeRemote(Remote *newrem, Displayable *dp)
	: Molecule(newrem->host(), dp) {
  do_construct(newrem);
}


// do main constructor tasks
void MoleculeRemote::do_construct(Remote *newrem) {
  rem = newrem;
  strName = strPath = NULL;
  Step = 0;
  hasPatchDisplay = FALSE;
  drawPatch = NULL;
  framesReceived = 0;
  MSGDEBUG(1,"Creating MoleculeRemote, host='" << rem->host() << "', app='");
  MSGDEBUG(1,rem->app() << "' ..." << sendmsg);
}


/////////////////////////  destructor  
MoleculeRemote::~MoleculeRemote(void) {
  MSGDEBUG(1,"Deleting MoleculeRemote ..." << sendmsg);
  
  if(strPath)  delete [] strPath;
  if(strName)  delete [] strName;
  
  // delete patch display, if possible
  if(drawPatch)
    delete drawPatch;

  // close remote connection
  if(rem)  delete rem;
}


/////////////////////  public routines  
// return the path of this molecule
char *MoleculeRemote::file_path(void) {
  if(!strPath)
    return rem->host();
  else
    return strPath;
}


// return the name of this molecule
char *MoleculeRemote::file_name(void) {
  if(!strName)
    return rem->app();
  else
    return strName;
}


// create the molecule from the data gained over the network.  return success.
int MoleculeRemote::create(void) {
  MSGDEBUG(2,"MoleculeRemote: creating structure ..." << sendmsg);
  
  if(nAtoms != 0) {
    msgErr << "Attempt to create molecule after it has already been created.";
    msgErr << sendmsg;
    return FALSE;
  }

  // make sure the remote simulation is connected ...
  if(!rem || ! rem->running_simulation())
    return FALSE;
    
  // we're here OK, so connection is good.  Get structure and create molecule.
  msgInfo << "Retrieving molecular structure from " << rem->host() << " ...";
  msgInfo << sendmsg;
  if(!(createFromRemote(rem)))
    return FALSE;

  // save the file name
  char *mname, *molname = rem->mol_name();
  if(molname) {
    mname = strrchr(molname, '/');
    if(!mname)
      mname = molname;
    else
      mname++;
  } else {
    mname = rem->host();
  }
  set_name(mname);

  // do anything else the parent class needs
  int retval = Molecule::create();
  
  // create patch representation object
  drawPatch = new DrawPatch(this);
  
  return retval;
}
  
  
// read in a molecule structure from the remote connection.
// Return TRUE if molecule was successfully read, FALSE if there is an error.
int MoleculeRemote::createFromRemote(Remote *rm) {
  int natoms, nbonds, i;
  char namesbuf[8], rnamesbuf[8], segnamesbuf[8], atypesbuf[8], ridstrbuf[8];
  char *names, *rnames, *segnames, *atypes, *ridstr;
  float newpos[ATOMCOORDS], newdata[ATOMEXTRA];
  struct md_struct *mds;

  // access molecular structure struct
  if(rm)
    mds = rm->md_struct;
  else
    return FALSE;
    
  natoms = mds->nAtoms;
  nbonds = mds->nBonds;
  
  MSGDEBUG(1,"MoleculeRemote: Creating structure from network (");
  MSGDEBUG(1,rem->host() << ") ..."  << sendmsg);

  // initialize data arrays that are used to create new atoms
  for(i=0; i < ATOMCOORDS; newpos[i++] = 0.0);
  namesbuf[mds->nameLen] = '\0';
  atypesbuf[mds->nameLen] = '\0';
  rnamesbuf[mds->nameLen] = '\0';
  segnamesbuf[mds->nameLen] = '\0';
  ridstrbuf[mds->nameLen] = '\0';

  MSGDEBUG(1,"                Remote structure with:" << sendmsg);
  MSGDEBUG(1,"                  " << natoms << " atoms" << sendmsg);
  MSGDEBUG(1,"                  " << nbonds << " bonds" << sendmsg);
  MSGDEBUG(1,"                  " << mds->nAtomNames << " atom names"<<sendmsg);
  MSGDEBUG(1,"                  " << mds->nAtomTypes << " atom types"<<sendmsg);

  // read in the atom structure info first
  int *typePos = mds->atomTypeIdx;
  int *namePos = mds->atomNameIdx;
  char *resIdPos = mds->resIds;
  char *resNamePos = mds->resNames;
  char *segNamePos = mds->segIds;
  float *massPos = mds->mass;
  float *chargePos = mds->charge;
  float *radiusPos = mds->radii;
  float *occupPos = mds->g;
  float *betaPos = mds->k;

  MSGDEBUG(2,"   Creating " << natoms << " atoms ..." << sendmsg);

  init_atoms(natoms);
  for(i = 0; i < natoms; i++) {
    names = namesbuf;
    atypes = atypesbuf;
    rnames = rnamesbuf;
    ridstr = ridstrbuf;
    segnames = segnamesbuf;

    // copy names into buffers, and increment pointers into mdcomm lists
    strncpy(names, mds->atomNameList + (mds->nameLen * *(namePos++)),
    	mds->nameLen);
    strncpy(atypes, mds->atomTypeList + (mds->nameLen * *(typePos++)),
        mds->nameLen);
    strncpy(rnames, resNamePos, mds->nameLen);
    strncpy(ridstr, resIdPos, mds->nameLen);
    strncpy(segnames, segNamePos, mds->nameLen);
    resNamePos += mds->nameLen;
    resIdPos += mds->nameLen;
    segNamePos += mds->nameLen;

    // save integer and float data
    newdata[ATOMCHARGE] = *(chargePos++);
    newdata[ATOMMASS] = *(massPos++);
    newdata[ATOMRAD] = *(radiusPos++);
    newdata[ATOMBETA] = *(betaPos++);
    newdata[ATOMOCCUP] = *(occupPos++);

    MSGDEBUG(1, "New atom " << i << ": name='" << names << "', mass=" 
		<< newdata[ATOMMASS] << sendmsg);

    add_atom(names, atypes, rnames, ridstr, segnames, newpos, newdata);
  }
  
  // now read bond data
  MSGDEBUG(2,"   Creating " << nbonds << " bonds ..." << sendmsg);

  // add bonds to current structure
  int *bondPos = mds->bonds;
  for(i=0; i < nbonds; i++) {
    add_bond(*bondPos, *(bondPos + 1));
    bondPos += 2;
  }

  MSGDEBUG(2,"Finished creating structure from remote connection." << sendmsg);

  return TRUE;
}


// function used by Displayable to prepare the Displayable object; here
//      it checks if new data is available from the remote machine
void MoleculeRemote::prepare(DisplayDevice *display) {
  Timestep *newts;
  Atom *atm;

  // make sure we're connected
  if(rem && nAtoms > 0) {

    MSGDEBUG(3,"Checking for new timestep from " << rem->host() << sendmsg);

    // check for new timestep
    if(rem->get_next_ts()) {

      MSGDEBUG(2,"New timestep " << *(rem->state.step) << " available from ");
      MSGDEBUG(2,rem->host() << sendmsg);

      // new step available - read it and add it to animation list
      // see if we need a new frame, or should replace the old one
      int saveFrame = (num() < 1 || framesReceived < 1 ||
      	(rem->saveFrameRate > 0 &&(framesReceived % rem->saveFrameRate == 0)));

      if(!saveFrame) {
        MSGDEBUG(3,"  Replacing frame " << num() - 1 << sendmsg);
        newts = item(num() - 1);
      } else {
        MSGDEBUG(3,"  Adding new frame " << num() << sendmsg);
        newts = new Timestep(nAtoms, 0.0);
      }

      // save new step number, and elapsed time from start of simulation
      Step = *(rem->state.step);
      newts->dt = *(rem->state.dt);

      // put coordinates into new timestep
      float *pos = newts->pos;
      float *tsx = rem->state.x;
      float *tsy = rem->state.y;
      float *tsz = rem->state.z;
      for(int i=0; i < nAtoms; i++) {
	if(num() == 0) {
	  // this is the first frame, put the positions in the Atom objects
	  atm = atom(i);
	  atm->pos[0] = *tsx;
	  atm->pos[1] = *tsy;
	  atm->pos[2] = *tsz;
	}
        *(pos++) = *(tsx++);
        *(pos++) = *(tsy++);
        *(pos++) = *(tsz++);
      }
      
      // put energies into new timestep
      newts->energy[TSE_BOND] = *(rem->state.ebond);
      newts->energy[TSE_ANGLE] = *(rem->state.eangle);
      newts->energy[TSE_DIHE] = *(rem->state.edihed);
      newts->energy[TSE_IMPR] = *(rem->state.eimpr);
      newts->energy[TSE_VDW] = *(rem->state.evdw);
      newts->energy[TSE_COUL] = *(rem->state.eelec);
      newts->energy[TSE_HBOND] = *(rem->state.ehbo);
      newts->energy[TSE_TEMP] = *(rem->state.temp);
      newts->energy[TSE_PE] = *(rem->state.epot);
      newts->energy[TSE_TOTAL] = *(rem->state.etot);
      newts->energy[TSE_KE] = *(rem->state.etot) - *(rem->state.epot);

      // save patch data, if available
      if(*(rem->state.nPatches) > 0) {
        float *pdata = newts->create_patch_storage(*(rem->state.nPatches));
        float *px = rem->state.patchX;
	float *py = rem->state.patchY;
	float *pz = rem->state.patchZ;
	float *pxlen = rem->state.patchXLen;
	float *pylen = rem->state.patchYLen;
	float *pzlen = rem->state.patchZLen;
	float *pload = rem->state.patchLoad;
	float *pnode = rem->state.patchNode;
	float *patoms = rem->state.patchAtoms;
        for(int np=0; np < newts->numPatches; np++) {
	  pdata[TSP_X] = *(px++);
	  pdata[TSP_Y] = *(py++);
	  pdata[TSP_Z] = *(pz++);
	  pdata[TSP_XLEN] = *(pxlen++);
	  pdata[TSP_YLEN] = *(pylen++);
	  pdata[TSP_ZLEN] = *(pzlen++);
	  pdata[TSP_LOAD] = *(pload++);
	  pdata[TSP_NODE] = *(pnode++);
	  pdata[TSP_ATOMS] = *(patoms++);
	  pdata += TSPATCHDATA;
	}
      } else {
        float *pdata = newts->create_patch_storage(1);
        pdata[TSP_X] = pdata[TSP_Y] = pdata[TSP_Z] = 0.0;
        pdata[TSP_XLEN] = pdata[TSP_YLEN] = pdata[TSP_ZLEN] = 0.0;
        pdata[TSP_LOAD] = 1.0;
        pdata[TSP_ATOMS] = (float)nAtoms;
        pdata[TSP_NODE] = 0.0;
      }

      if(newts->numPatches > 0)
        hasPatchDisplay = TRUE;

      // signal that a new timestep can be sent
      rem->done_with_ts();

      // now append the new frame, if necessary
      if(saveFrame) {
        append_frame(newts);		// append the new frame
      } else {
        newts->init();			// recalc values for the timestep
	curr_frame_changed();		// signal the current frame changed
      }
      framesReceived++;

      // change patch loc and size JUST FOR THE CASE WHERE NO PATCH DATA
      // GIVEN IN MDCOMM DATA
      // this makes the single patch the size of the current frame
      if(*(rem->state.nPatches) < 1) {
        float *pdata = newts->patchData;
	pdata[TSP_X] = newts->minpos[0];
	pdata[TSP_Y] = newts->minpos[1];
	pdata[TSP_Z] = newts->minpos[2];
	pdata[TSP_XLEN] = newts->maxpos[0] - pdata[TSP_X];
	pdata[TSP_YLEN] = newts->maxpos[1] - pdata[TSP_Y];
	pdata[TSP_ZLEN] = newts->maxpos[2] - pdata[TSP_Z];
	newts->minpatch[TSP_X] = pdata[TSP_X];
	newts->minpatch[TSP_Y] = pdata[TSP_Y];
	newts->minpatch[TSP_Z] = pdata[TSP_Z];
	newts->maxpatch[TSP_X] = pdata[TSP_X] + pdata[TSP_XLEN];
	newts->maxpatch[TSP_Y] = pdata[TSP_Y] + pdata[TSP_YLEN];
	newts->maxpatch[TSP_Z] = pdata[TSP_Z] + pdata[TSP_ZLEN];
	newts->minpatch[TSP_XLEN] = newts->maxpatch[TSP_XLEN] =pdata[TSP_XLEN];
	newts->minpatch[TSP_YLEN] = newts->maxpatch[TSP_YLEN] =pdata[TSP_YLEN];
	newts->minpatch[TSP_ZLEN] = newts->maxpatch[TSP_ZLEN] =pdata[TSP_ZLEN];
      }
    }
  }
      
  // do any other preparations necessary
  Molecule::prepare(display);
}


// get the method for displaying patches
int MoleculeRemote::patch_display_method(void) {
  if(drawPatch)
    return drawPatch->display_method();
  else
    return DrawPatch::BYNONE;
}
  
  
// change the method for displaying patches
int MoleculeRemote::patch_display_method(int newval) {
  if(drawPatch)
    return drawPatch->display_method(newval);
  else
    return DrawPatch::BYNONE;
}

